"""The ``push`` sub-package contains an extensible Push interpreter and Push program abstraction.

When running Push programs (either during search or after) the modules in this package are used to perform the
stack-based execution of Push code and report output values. For more information on the Push programming language,
see the following links

- http://faculty.hampshire.edu/lspector/push3-description.html
- https://erp12.github.io/push-redux/

"""
